/**
 * test-webid
 * Copyright (C) 2011 Thomas Bergwinkl
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package org.bergnet.test.webid.testcases;

import java.io.File;
import java.io.FileWriter;
import java.io.StringReader;
import java.lang.reflect.Method;
import java.net.URL;

import org.apache.http.Header;
import org.apache.http.HttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.util.EntityUtils;
import org.bergnet.test.EarlTestCase;
import org.bergnet.test.webid.WebIDTestException;
import org.bergnet.test.webid.WebIDTests;
import org.bergnet.test.webid.XsdAnyUriDatatype;
import org.bergnet.webid.commons.CryptoUtils;
import org.bergnet.webid.commons.FOAFBuilder;
import org.bergnet.webid.commons.WebID;
import org.bergnet.webid.commons.WebIDHttpClient;
import org.junit.Test;

import com.hp.hpl.jena.query.Query;
import com.hp.hpl.jena.query.QueryExecution;
import com.hp.hpl.jena.query.QueryExecutionFactory;
import com.hp.hpl.jena.query.QueryFactory;
import com.hp.hpl.jena.query.QuerySolutionMap;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ResourceFactory;


public abstract class AbstractTestCase {
	public static final String PROPERTY_KEY_SIZE = "org.bergnet.commons.test.webid.keySize";

	protected int keySize;
	protected URL endpoint;

	protected String id = null;
	protected String fragment = "#me";

	protected WebID webid = null;
	protected String foaf = null;
	protected URL baseUrl = null;
	protected String output = null;
	protected String outputFormat = null;
	protected String verificationQuery = null;
	protected QuerySolutionMap verificationVariables = new QuerySolutionMap();
	protected boolean result = false;

	protected void init() throws WebIDTestException {
		String keySizeString = WebIDTests.getInstance().getConfig().getProperties().getProperty(PROPERTY_KEY_SIZE);

		if(keySizeString == null)
			throw new WebIDTestException("Key size not defined");

		keySize = Integer.parseInt(keySizeString);

		endpoint = WebIDTests.getInstance().getConfig().getEndpoint();

		if(id != null) {
			String webid = getFullUrl(id + fragment);

			verificationVariables.add("webid", ResourceFactory.createResource(webid));
			verificationVariables.add("webidString", ResourceFactory.createTypedLiteral(webid, XsdAnyUriDatatype.getInstance()));
		}
	}

	public WebID generateWebID() throws WebIDTestException {
		try {
			webid = WebID.generate(keySize, getFullUrl(id + fragment), WebID.createDeltaDate(-1), WebID.createDeltaDate(1));;

			CryptoUtils.writePemFile(webid, getOutputFolder() + "/webid.pem");

			return webid; 
		} catch(Exception e) {
			throw new WebIDTestException("Can't generate WebID", e);
		}
	}

	public String buildFOAF() throws WebIDTestException {
		try {
			foaf = FOAFBuilder.build(webid, FOAFBuilder.FORMAT_RDF_XML);

			dump("foaf.rdf", foaf);

			return foaf;
		} catch(Exception e) {
			throw new WebIDTestException("Can't build FOAF file", e);
		}
	}

	public void publishFOAF() throws WebIDTestException {
		WebIDTests.getInstance().getPublisher().publish(id, foaf, FOAFBuilder.FORMAT_RDF_XML);
	}

	public String login() throws WebIDTestException {
		try {
			int port = (endpoint.getPort() != -1 ? endpoint.getPort() : 443);

			WebIDHttpClient httpClient = new WebIDHttpClient(webid, port);

			try {
				HttpGet httpGet = new HttpGet(endpoint.toString());
				httpGet.setHeader("Accept", "application/rdf+xml,text/turtle,text/n3;q=0.9,*/*;q=0.1");

				HttpResponse response = httpClient.execute(httpGet);

				outputFormat = "RDF/XML";

				for(Header contentType : response.getHeaders("Content-Type")) {
					if("text/n3".equals(contentType.getValue()))
						outputFormat = "N3";
					else if("text/turtle".equals(contentType.getValue()))
						outputFormat = "TTL";
				}

				// TODO: handle redirects
				baseUrl = endpoint;

				output = EntityUtils.toString(response.getEntity());
			} finally {
				httpClient.getConnectionManager().shutdown();
			}

			dump("response.txt", output);		
		} catch(Exception e) {
			throw new WebIDTestException("Error during login", e);
		}

		return output;
	}

	public boolean verifyOutput() throws WebIDTestException {
		try {
			Model model = ModelFactory.createDefaultModel();
			model.read(new StringReader(output), baseUrl.toString(), outputFormat);

			Query query = QueryFactory.create(verificationQuery);
			
			QueryExecution queryExecution = QueryExecutionFactory.create(query, model, verificationVariables);

			return result = queryExecution.execAsk();
		} catch(Exception e) {
			throw new WebIDTestException("Can't verify output", e);
		}
	}

	protected String getFullUrl(String file) throws WebIDTestException {
		 return WebIDTests.getInstance().getPublisher().getBaseUrl() + file;
	}

	protected String getTestCaseShortName() throws WebIDTestException {
		try {
			return new URL(getEarlTestCase().url()).getRef();
		} catch(Exception e) {
			throw new WebIDTestException("Can't find test case short name", e);
		}
	}

	protected String getOutputFolder() throws WebIDTestException {
		String baseOutputFolder = WebIDTests.getInstance().getOutputFolder();
		String shortName = getTestCaseShortName();
		String outputFolder = baseOutputFolder + "/" + shortName;

		File outputFolderFile = new File(outputFolder);

		if(!outputFolderFile.exists())
			outputFolderFile.mkdir();

		return outputFolder;
	}

	protected void dump(String fileName, String content) throws WebIDTestException {
		try {
			String outputFolder = getOutputFolder();

			FileWriter writer = new FileWriter(outputFolder + "/" + fileName);
			writer.write(content);
			writer.close();
		} catch(Exception e) {
			throw new WebIDTestException("Can't dump debug file", e);
		}
	}

	private EarlTestCase getEarlTestCase() {
		Method method = getTestMethod();

		if(method != null && method.isAnnotationPresent(EarlTestCase.class))
			return method.getAnnotation(EarlTestCase.class);

		return null;
	}

	private Method getTestMethod() {
		int index = 0; 

		StackTraceElement[] stackTraceElements = Thread.currentThread().getStackTrace();

		while(index < stackTraceElements.length && !isTestMethod(stackTraceElements[index]))
			index++;

		String className = stackTraceElements[index].getClassName();
		String methodName = stackTraceElements[index].getMethodName();

		try {
			return Class.forName(className).getMethod(methodName);
		} catch(Exception e) {}

		return null;
	}

	private boolean isTestMethod(String className, String methodName) {
		try {
			Method method = Class.forName(className).getMethod(methodName);

			return method.isAnnotationPresent(Test.class);
		} catch(Exception e) {}

		return false;
	}

	private boolean isTestMethod(StackTraceElement stackTraceElement) {
		return isTestMethod(stackTraceElement.getClassName(), stackTraceElement.getMethodName());
	}
}
